#!/usr/bin/php
<?php
/*
 * © Copyright 2007, 2008 IntraHealth International, Inc.
 * 
 * This File is part of iHRIS
 * 
 * iHRIS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * The page wrangler
 * 
 * This page loads the main HTML template for the home page of the site.
 * @package iHRIS
 * @subpackage DemoManage
 * @access public
 * @author Carl Leitner <litlfred@ibiblio.org>
 * @copyright Copyright &copy; 2007, 2008 IntraHealth International, Inc. 
 * @since Demo-v2.a
 * @version Demo-v2.a
 */



$dir = getcwd();
chdir("../pages");
$i2ce_site_user_access_init = null;
$wd = getcwd();

$i2ce_site_user_database = null;
require_once( $wd . DIRECTORY_SEPARATOR . 'config.values.php');

$local_config = $wd . DIRECTORY_SEPARATOR .'local' . DIRECTORY_SEPARATOR . 'config.values.php';
if (file_exists($local_config)) {
    require_once($local_config);
}

if(!isset($i2ce_site_i2ce_path) || !is_dir($i2ce_site_i2ce_path)) {
    echo "Please set the \$i2ce_site_i2ce_path in $local_config";
    exit(55);
}

require_once ($i2ce_site_i2ce_path . DIRECTORY_SEPARATOR . 'I2CE_config.inc.php');
I2CE::raiseError("Connecting to DB");
if (isset($i2ce_site_dsn)) {
    @I2CE::initializeDSN($i2ce_site_dsn,   $i2ce_site_user_access_init,    $i2ce_site_module_config);         
} else if (isset($i2ce_site_database_user)) {    
    I2CE::initialize($i2ce_site_database_user,
                     $i2ce_site_database_password,
                     $i2ce_site_database,
                     $i2ce_site_user_database,
                     $i2ce_site_module_config         
        );
} else {
    die("Do not know how to configure system\n");
}

I2CE::raiseError("Connected to DB");
$db= MDB2::singleton();
$ff = I2CE_FormFactory::instance();
$user = new I2CE_User();


$sql = "CREATE TABLE IF NOT EXISTS `infinium_import` (`hash` BINARY(16) NOT NULL, UNIQUE KEY `hash` (`hash`))";
if (I2CE::pearError($db->exec($sql),"Cannot create logging table", E_USER_ERROR)) {
    die();
}


require_once($i2ce_site_i2ce_path . DIRECTORY_SEPARATOR . 'tools' . DIRECTORY_SEPARATOR . 'CLI.php');



require_once($i2ce_site_i2ce_path . DIRECTORY_SEPARATOR . 'tools' . DIRECTORY_SEPARATOR . 'import_base.php');


$expected_headers = array(
    'ministry'=>'Ministry',
    'emp_num'=>'Employee Number',
    'omang'=>'Omang/ Exemption',
    'date'=>'Effective Date',
    'transaction'=>'Transaction Type',
    'dpsm_code'=>'New Position',
    'new_pos_title'=>'New Position Title',
    'new_loc'=>'New Location',
    'prev_pos'=>'Previous Position',
    'prev_pos_title'=>'Previous Position Title',
    'prev_code'=>'Previous Location',
    'term_code'=>'Termination Code',
    'term_desc'=>'Termination Description',
    'new_stat_code'=>'New Status Code',
    'new_stat_desc'=>'New Status Description',
    'prev_stat_code'=>'Previous Status Code',
    'prev_stat_desc'=>'Previous Status Description',
    'prev_surnname'=>'Prev Surname',
    'prev_forename'=>'Prev. Forename',
    'surname'=>'Surname',
    'forename'=>'Forename');
$required_cols_by_transaction = array(
    'EE'=>array('omang','date','term_code','new_stat_code'),
    'NE'=>array('omang','date','surname','forename','new_stat_code','dpsm_code','new_pos_title'),
    'PS'=>array('omang','date','dpsm_code','new_pos_title'),
    'ST'=>array('omang','date','new_stat_code')
    );

$infinium_termination_map = array(
    'DISMI'=>'pos_change_reason|3', //dismisal
    'DEATH'=>'pos_change_reason|1' //death
    // 'CONTE'=>'', //terminaion of contract
    // 'COMRE'=>'', //Compulory retirement
    // 'VOLRE'=>'', //volunary retirement
    // 'EXTTR'=>'', //transfer to another ministry
    // 'PURGE'=>'', //code to purge employees  --- SKIP!
    // 'ENDCO'=>'' //end of contract
    );
$infinium_stat_code_map = array(
    'TEMPF'=>'position_type|8',//Temp full time -- double check
    'PERM'=>'position_type|1',//Permanent and penshionable
    'FTCON'=>'position_type|2',//fixed term contract --douvle check
    'PROB'=>'position_type|6' //probabionary
    //nothing for acting.
    );



if (count($arg_files) != 1) {
    usage("Please specify the name of a .CSV file to import");
}

reset($arg_files);
$file = current($arg_files);
if($file[0] == '/') {
    $file = realpath(current($arg_files));
} else {
    $file = realpath($dir. '/' . $file);
}

if (!is_readable($file) || ($fp = fopen($file,"r")) === false) {
    usage("Please specify the name of a .CSV spreadsheet to import: " . $file . " is not readable");
}

$headers = loadHeadersFromCSV($fp);
mapHeaders($headers);

$info = pathinfo($file);
$bad_file_name = dirname($file) . DIRECTORY_SEPARATOR . basename($file,'.'.$info['extension']) . '.bad_' .date('d-m-Y') .'.csv';
I2CE::raiseError("Will put any bad records in $bad_file_name");

$bad_headers = $headers;
$bad_headers[] = "Row In " . basename($file);
$bad_headers[] = "Reasons for failure";


loadDPSMData();

$row = 0;
while (($data = fgetcsv($fp)) !== FALSE) {
    $row++;
    processTransaction($data,$row);    
}

/*********************************************
*
*      Helper functions
*
*********************************************/

function loadHeadersFromCSV($fp) {
    $in_file_sep = false;
    foreach (array("\t",",") as $sep) {
        $headers = fgetcsv($fp, 4000, $sep);
        if ( $headers === FALSE|| count($headers) < 3) {
            fseek($fp,0);
            continue;
        }
        $in_file_sep = $sep;
    }
    if (!$in_file_sep) {
        die("Could not get headers\n");
    }
    foreach ($headers as &$header) {
        $header = trim($header);
    }
    unset($header);
    return $headers;
}

function mapHeaders($headers) {
    global $expected_headers;
    global $header_map = array();
    foreach ($expected_headers as $expected_header_ref => $expected_header) {
        if (($header_col = array_search($expected_header,$headers)) === false) {
            I2CE::raiseError("Could not find $expected header in the following headers:\n\t" . implode(" ", $expected_headers). "\nFull list of found headers is:\n\t" . implode(" ", $headers));
            die();
        }
        $header_map[$expected_header_ref] = $header_col;
    }
}


function loadDPSMData() {
    global $dpsm_code_map = array();
    global $dpsm_grade = array();
    $salary_grades = I2CE_FormStorage::listFields('salary_grade','name');
    foreach ($salary_grades as $id=>&$data) {
        if (!is_array($data) || !array_key_exists('name',$data) || !$data['name']) {
            unset($salary_grades[$id]);
            continue;
        }
        $data = $data['name'];
    }
    unset($data);
    foreach (I2CE_FormStorage::listFields('post',array('dpsm_code','salary_grade')) as $id=>$data) {
        if (!is_array($data) || !array_key_exists('dpsm_code',$data) || !$data['dpsm_code']) {
            continue;
        }
        $dpsm_code_map[strtoupper($data['dpsm_code'])] =  $id;
        $grade = false;
        if (array_key_exists('salary_grade',$data) && $data['salary_grade'] && count($grades = explode(",",$data['salary_grade'])) == 1) {
            $salary_grade_id = substr($grades[0],14);
            if (is_string($salary_grade_id) && strlen($salary_grade_id) > 0 && array_key_exists($salary_grade_id,$salary_grades)) {
                $grade = $salary_grades[$salary_grade_id];
            }
        }
        $dpsm_grade[strtoupper($data['dpsm_code'])] = $grade;
    }
}

function addBadRecord($raw_data,$row,$reason) {
    global $bad_fp;
    global $bad_file_name;
    global $bad_headers;
    if (!is_resource($bad_fp)) {
        $bad_fp = fopen($bad_file_name,"w");
        if (!is_resource($bad_fp)) {
            I2CE::raiseError("Could not open $bad_file_name for writing.", E_USER_ERROR);
            die();
        }        
        fputcsv($bad_fp, $bad_headers);
    }
    I2CE::raiseError("Skipping processing of row $row: $reason");
    $raw_data[] = $row;
    $raw_data[] = $reason;
    fputcsv($bad_fp, $raw_data);
}
 
function findPerson($omang) {
    $omang_id_type = 'id_type|1';
    $exemption_id_type = 'id_type|2';
    $omang = strtoupper(trim($omang));
    if (ctype_digit($omang)) {
        $id_type = $omang_id_type;
    }else if (in_array($omang, array('EXP', 'EXPATRIATE'))) {
        return false; //expatriate
    } else {
        $id_type = $excemption_id_type;
        //exception id
    }
    $where = array(
        'operator'=>'AND',
        'operand'=>array(
            0=>array(
                'operator'=>'FIELD_LIMIT',
                'field'=>'id_type',
                'style'=>'equals',
                'data'=>array(
                    'value'=>$id_type
                    )
                ),
            1=>array(
                'operator'=>'FIELD_LIMIT',
                'field'=>'id_num',
                'style'=>'equals',
                'data'=>array(
                    'value'=>$omang
                    )
                )
            )
        );
    $persons = I2CE_FormStorage::listFields('person_id',array('parent'),true,$where);
    $person_id = false;
    if (count($persons) != 1) {
        return false;
    }
    $data = current($persons);
    if (array_key_exists('parent',$data) && is_string($data['parent']) && substr($data['parent'],0,7) == 'person|') {
        $person_id = substr($data['parent'],7);
    }            
    if (!$person_id) {
        return false;
    }
    $personObj = $form_factory->createForm('person|' . $person_id);
    if (!$personObj instanceof iHRIS_Person) {
        return false;
    }
    $personObj->populate();
    return $personObj;
}

function findLastPosition($omang,$only_current) {
    if (! ($personObj = findPerson($omang)) instanceof iHRIS_Person) {
        return false;
    }
    if ($only_current) {
        $where = array(
            'operator'=>'FIELD_LIMIT',
            'field'=>'end_date',
            'style'=>'null'
            );
    } else {
        $where = array();
    }
    $per_pos_id = I2CE_FormStorage::search('person_position', $personObj->getNameId(),$where,'-start_date',1);
    if (!$per_pos_id) {
        return false;
    }
    $persPosObj = I2CE_FormFactory::instance()->createContainer('person_position'.'|'.$per_pos_id);
    if (!$persPosObj instanceof iHRIS_PersonPosition) {
        return false;
    }
    $persPosObj->populate();
    return $persPosObj;
}


function mapData($data,$row) {
    global $header_map;
    $mapped_data = array();
    foreach ($header_map as $header_ref=>$header_col) {
        if (!array_key_exists($header_col,$data)) {
            $mapped_data[$header_ref] = false;
        } else {
            $mapped_data[$header_ref] = $data[$header_col];
        }
    }
    
    if (!$mapped_data['transaction']) {
        addBadRecord($data,$row,'No transcation type');
        return false;
    }
    return $mapped_data;
}

function verifyData($data,$row,$transaction) {
    global $required_cols_by_transaction;    
    if (!array_key_exists($transaction,$required_cols_by_transaction)) {
        addBadRecord($data,$row,"Invalid transaction type");
        return false;
    }
    $missing_cols = array();
    foreach ($required_cols_by_transaction[$transaction] as $required_col) {
        if ($mapped_data[$required_col] === false || (is_string($mapped_data[$required_col]) && strlen($mapped_data[$required_col]) == 0)) {
            $missing_cols[] = $required_col;
        }
    }
    if (count($missing_cols) > 0) {
        addBadRecord($data,$row,"Missing required columns " . implode(" ",$missing_cols));
        return false;
    }
    return true;
}

function getDate($data,$row,$date) {
    //first check the date e.g 16/05/2011
    $mateches = array();
    if (!preg_match('/^([0-9]+\)\\/([0-9]+)\\/([0-9]+)$/'.trim($date),$matches)) {
        addBadRecord($data,$row,"Bad date format");
        return false;
    }
    list($date,$day,$month,$year) = $matches;
    if ($day < 1 || $day > 31 || $month < 1 || $month > 12 || $year < 1900 || $year > 2100) {
        addBadRecord($data,$row,"Invalid date");
        return false;
    }
    return I2CE_DATE::getDate($day,$month,$year);
}


function alreadyProcessed($data,$row,$mapped_data) {
    global $expected_headers;
    global $db;
    $hash_data = '';
    foreach (array_keys($expected_headers) as $header_ref) {
        $hash_data .= $mapped_data[$header_ref];
    }
    $hash_val = mysql_real_escape_string(md5($hash_data,true));
    $hash_check = 'SELECT hash FROM infinium_import WHERE hash = \''. $hash_val  .'\'';
    if (I2CE::pearError(($res = $db->query($hash_check)),"Cannot check value in logging table", E_USER_ERROR)) {
        die();
    }
    if ( $res->fetchRow()) {
        addBadRecord($data,$row,'Transaction had already been processed');
        return false;
    }
    return $hash_val;
}

function markProcessed($hash_val) {
    global $db;
    $hash_insert = 'INSERT into  infinium_import VALUES (\''.  $hash_val .'\')';
    if (I2CE::pearError($db->exec($hash_insert),"Cannot add value to logging table", E_USER_ERROR)) {
        die();
    }
    
}


function processEmployeeExit($data,$row,$mapped_data) {
    global $infinium_termination_map;
    $ff = I2CE_FormFactory::instance();
    //first If person does not exist, error out
    if (! ($personObj = findPerson($mapped_data['omang'])) instanceof iHRIS_Person) {
        addBadRecord("Person not found");
        return false;
    }
    if (! ($personPosObj = findCurrentPosition($mapped_data['omang'],true)) instanceof iHRIS_PersonPosition) {
        addBadRecord("No current position found for this person");
        $personObj->cleanup();
        return false;
    }

    $pos_change_reason = false;
    if (array_key_exists($mapped_data['term_code'],$infinium_termination_map)) {
        $pos_change_reason = $infinium_termination_map[$mapped_data['term_code']];
    } else {
        $pos_change_reason = addTerminiation($mapped_data['term_code'],$mapped_data['term_desc']);
    }
    if (!$pos_change_reason) {
        addBadRecord("Termination code has not been mapped to iHRIS");
        $personPosObj->cleanup();
        $personObj->cleanup();
        return false;
    }
    if (! ($posObj = $ff->createContainer( $personPosObj->position))instanceof iHRIS_Position) {
        addBadRecord("No position could be created");
        $personPosObj->cleanup();
        $personObj->cleanup();
        return false;
    }
    $posObj->populate();

    $persPosObj->end_date = $effective_date;
    $persPosObj->reason = explode("|",$pos_change_reason);
    $pos_status_id = 'discontinued'; //also 'closed' or 'discontinued' as possibilities
    $posObj->status = array('position_status',$pos_status_id);
    $personPosObj->save($user);
    $posObj->save($user);
    $personPosObj->cleanup();
    $posObj->cleanup();
    $personObj->cleanup();
    return true;        
}

function processStatusChange($data,$row,$mapped_data) {
    global $infinium_stat_code_map;
    $ff = I2CE_FormFactory::instance();
    if (! ($existingPersonPosObj = findCurrentPosition($mapped_data['omang'],true)) instanceof iHRIS_PersonPosition) {
        addBadRecord("No current position found for this person");
        return false;
    }
    if (array_key_exists($mapped_data['new_stat_code'],$infinium_stat_code_map)) {
        $new_stat_code = $infinium_stat_code_map[$mapped_data['new_stat_code']];
    } else {
        $new_stat_code = addStatCode($mapped_data['new_stat_code'],$mapped_data['new_stat_desc']);
    }
    if (!$new_stat_code ) {
        addBadRecord("New Position Status Code has not been mapped to iHRIS");
        $existingPersonPosObj->cleanup();
        return false;
    }
    if (! ($existingPosObj = $ff->createContainer( $existingPersonPosObj->position))instanceof iHRIS_Position) {
        addBadRecord("No position could be created");
        $existingPersonPosObj->cleanup();
        return false;
    }
    $existingPosObj->populate();
    $existingPosObj->pos_type = array('position_type',$new_stat_code);
    $existingPosObj->save($user);
    $existingPosObj->cleanup();
    return true;
}


function createNewPerson($data,$row,$mapped_data) {
    $personObj = findPerson($mapped_data['omang']);
    if ($personObj instanceof iHRIS_Person) {
        addBadRecord("Person already exists in system with that OMANGE number");
        return false;
    }
    //for a NE we create the person
    $personObj = $ff->createContainer('person');
    $personObj->firstname = $mapped_data['forename'];
    $personObj->surname = $mapped_data['surname'];
    return $personObj;
}

function closeAttachedPosition($data,$row,$existingPersonPosObj,$effective_date) {
    if (! ($existingPosObj = $ff->createContainer( $existingPersonPosObj->position))instanceof iHRIS_Position) {
        addBadRecord($data,$row,"No position could be created");
        return false;
    }
    $existingPosObj->populate();



    $existingPersPosObj->end_date = $effective_date;
    //$existingPersPosObj->reason = explode("|",$pos_change_reason);
    $existingPosObj->status = array('position_status','discontinued');
    $existingPersonPosObj->save($user);
    $existingPosObj->save($user);
    $existingPosObj->cleanup();
    return true;
}



function processTransaction($data,$row) {
    if ( ! ($mapped_data = mapData($data,$row))) {
        return false;
    }
    if (!verifyData($data,$row,$mapped_data['transaction'])) {
        return false;
    }
    if (($hash = alreadyProcessed($data,$row,$mapped_data)) === true) {
        return false;
    }
    if ( ! ($effective_date = getDate($data,$row,$mapped_data['date']))) {
        return false;
    }
    $success = false;
    switch ($mapped_data['transaction']) {
    case 'EE':
        $success = processEmployeeExit($data,$row,$mapped_data);
        break;
    case 'ST':
        $success = processStatusChange($data,$row,$mapped_data);
        break;
    case 'PS': 
        if (! ($personObj = findPerson($mapped_data['omang'])) instanceof iHRIS_Person) {
            addBadRecord($data,$row,"Person not found");
            break;
        }
        $new_state_code = false;
        if ( ($existingPersonPosObj = findCurrentPosition($mapped_data['omang'],true)) instanceof iHRIS_PersonPosition) {
            $new_stat_code = existingPOSOBJ+position_type;            
            $closed = closeAttachedPosition($existingPersonPosObj,$effective_date);
            $existingPersonPosObj->cleanup();
            if (!$closed) {
                $personObj->cleanup();
                break;
            }
        }
        $success = setNewPostion($data,$row,$personObj,$mapped_data,$effective_date,$new_stat_code);
        $personObj->cleanup();
        break;
    case 'NE':
        if (! ($personObj = createNewPerson($data,$row,$mapped_data))) {
            break;
        }
        if (array_key_exists($mapped_data['new_stat_code'],$infinium_stat_code_map)) {
            $new_stat_code = $infinium_stat_code_map[$mapped_data['new_stat_code']];
        } else {
            $new_stat_code = addStatCode($mapped_data['new_stat_code'],$mapped_data['new_stat_desc']);
        }
        if (!$new_stat_code ) {
            addBadRecord("New Position Status Code has not been mapped to iHRIS");
            $personObj->cleanup();
            break;
        }
        $success = setNewPostion($data,$row,$personObj,$mapped_data,$effective_date,$new_stat_code);
        $personObj->cleanup();
        break;
    default:
        addBadRecord($data,$row,"Do not know how to handle the transaction");
        break;
    }
    if (!$success) {
        return false;
    }
    //if we make it here then the record has been successfuly processed
    markProcessed($hash_val);    
    return true;
}

function lookupFacility($mapped_data) {
    return false;
}

function setNewPosition($data,$row,$personObj,$mapped_data,$effective_date,$new_stat_code) {
    global $dpsm_code_map;
    global $dpsm_grade;
    $postid = array_search($mapped_data['dpsm_code'],$dpsm_code_map);
    if ($postid === false) {
        addBadRecord($data,$row,"DPSM Code is not in system");
        return false;
    }
    //create new forms
    $newPosObj = $ff->createContainer('position');
    $newPersonPosObj = $ff->createContainer('person_position');
    $newSalaryObj = $ff->createContainer('salary');


    //post+dpms_code shoud be the new position code
    //set new form data
    $newPosObj->status = array('position_status','closed');
    $newPosObj->post = array('post',$postid);
    if ($new_stat_code  !== false) {
        $newPosObj->pos_type = array('position_type',$new_stat_code);
    }
    $newPosObj->title = $mapped_data['new_pos_title'];
    if ( ($facilityid = lookupFacility($mapped_data)) !== false) {
        $newPosObj->facility = array('facility',$facilityid);
    }
    $newPosObj->save($user);


    $newPersonPosObj->start_date = $effective_date;
    $newPersonPosObj->position =  array('position',$newPosObj->getId());    
    $newPersonPosObj->setParent($personObj);
    $newPersonPosObj->save($user);


    $newSalaryObj->start_date = $effective_date;
    // post+salary_grade is a map_mult .  if it has exactly one selection this shold be the salarycode
    $salarycode = false;
    $salarycode = $dpsm_grade[$mapped_data['dpsm_code']];
    if ($salarycode) {
        $newSalayObj->salary = array('currency','1',$salarycode);  //currency|1 is BWP
    }
    $newSalaryObj->setParent($newPersonPosObj);
    $newSalaryObj->save($user);

    $newPersonPosObj->cleanup();
    $newSalaryObj->cleanup();
    $newPosObj->cleanup();
    return true;
}





function addTermination($code,$title) {                                                
    global $infininium_termination_map;
    if (!$title) {
        return false;
    }
    if (strtoupper($code) == 'PURGE') {
        return false;
    }
    $existing_terminations = I2CE_FormStorage::listFields('pos_change_reason','name');
    foreach ($existing_terminations as $id=>&$data) {
        if (!is_array($data) || !array_key_exists('name',$data) || !$data['name']) {
            unset($existing_terminations[$id]);
            continue;
        }
        $data = $data['name'];
    }
    unset($data);

    if ( ($id = array_search($title,$existing_terminations))!==false) {
        return "pos_change_reason|" . $id;
    }
    if (!simple_prompt("Could not find position change reason $code: $title in " . implode(",",$existing_terminations) . "Should I add it?")) {
        return false;
    }
    $posChangeObj = $ff->createContainer('position_type');
    if (!$posChangeObj instanceof I2CE_SimpleList) {
        return false;
    }
    $posChangeObj->setID($code);
    $posChangeObj->getField('name')->setValue($title);    
    $posChangeObj->save($user);
    $posChangeObj->cleanup();
    $mapped = 'pos_change_reason|' . $code;
    $infinium_termination_map[$code] = $mapped;
    return $mapped;
}
    

function  addStatCode($code,$title) {
    global $infininium_stat_code_map;
    if (!$title) {
        return false;
    }
    $existing_stat_codes = I2CE_FormStorage::listFields('position_type','name');
    foreach ($existing_stat_codes as $id=>&$data) {
        if (!is_array($data) || !array_key_exists('name',$data) || !$data['name']) {
            unset($existing_stat_codes[$id]);
            continue;
        }
        $data = $data['name'];
    }
    unset($data);

    if ( ($id = array_search($title,$existing_stat_codes))!==false) {
        return "position_type|" . $id;
    }
    if (!simple_prompt("Could not find position type $code: $title in " . implode(",",$existing_stat_codes) . "Should I add it?")) {
        return false;
    }
    $posTypeObj = $ff->createContainer('position_type');
    if (!$posTypeObj instanceof I2CE_SimpleList) {
        return false;
    }
    $posTypeObj->setID($code);
    $posTypeObj->getField('name')->setValue($title);    
    $posTypeObj->save($user);
    $posTypeObj->cleanup();
    $mapped = 'position_type|' . $code;
    $infinium_stat_code_map[$code] = $mapped;
    return $mapped;
        
}





# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
